import logging
from datetime import datetime
from typing import List, Optional

from ...enums import MatchLevel
from .constants import MatchReportInfoSectionFields
from .fields import parse_match_date, parse_match_level
from .models import ParsedMatchInfo

_logger = logging.getLogger(__name__)


MATCH_REPORT_PREFIX_NAME = f"{MatchReportInfoSectionFields.MATCH_NAME.value}:"
MATCH_REPORT_PREFIX_DATE = f"{MatchReportInfoSectionFields.MATCH_DATE.value}:"
MATCH_REPORT_PREFIX_LEVEL = f"{MatchReportInfoSectionFields.MATCH_LEVEL.value}:"


def parse_match_info(info_lines: List[str]) -> ParsedMatchInfo:
    name: Optional[str] = None
    raw_date: Optional[str] = None
    date: Optional[datetime] = None
    level: Optional[MatchLevel] = None
    for line in info_lines:
        if line.startswith(MATCH_REPORT_PREFIX_NAME):
            name = line.partition(MATCH_REPORT_PREFIX_NAME)[-1]
            _logger.debug("match name found: %s", name)
        elif line.startswith(MATCH_REPORT_PREFIX_DATE):
            raw_date = line.partition(MATCH_REPORT_PREFIX_DATE)[-1]
            _logger.debug("match date found: %s", raw_date)
            date = parse_match_date(raw_date)
        elif line.startswith(MATCH_REPORT_PREFIX_LEVEL):
            level_text = line.partition(MATCH_REPORT_PREFIX_LEVEL)[-1].upper()
            _logger.debug("attempting to parse match level: %s", level_text)
            level = parse_match_level(level_text)
            _logger.debug("parsed match level: %s", level)
    return ParsedMatchInfo(name=name, raw_date=raw_date, date=date, match_level=level)



import logging
import re
from datetime import datetime
from typing import Any, Optional

from ...enums import Classification, Division, MatchLevel, PowerFactor, Scoring

_logger = logging.getLogger(__name__)


def parse_match_level(level_text: str) -> Optional[MatchLevel]:
    if not level_text:
        return None
    if "IV" in level_text:
        return MatchLevel.IV
    if "III" in level_text:
        return MatchLevel.III
    if "II" in level_text:
        return MatchLevel.II
    if "I" in level_text:
        return MatchLevel.I
    return None


MATCH_REPORT_DATE_FORMAT = "%m/%d/%Y"


def parse_match_date(date_text: str) -> Optional[datetime]:
    try:
        parsed_date = datetime.strptime(date_text, MATCH_REPORT_DATE_FORMAT)
        return parsed_date
    except (ValueError, TypeError):
        _logger.debug("failed to parse match date: %s", date_text)
        return None


def _sanitize_string(s: Any) -> str:
    return (s if (s and isinstance(s, str)) else "").strip()


def parse_division(s: str) -> Division:
    match _sanitize_string(s).lower():
        case "pcc" | "pistol caliber carbine":
            return Division.PCC
        case "open":
            return Division.OPEN
        case "ltd" | "limited":
            return Division.LIMITED
        case "carry optics" | "co":
            return Division.CARRY_OPTICS
        case "l10" | "ltd10" | "limited 10":
            return Division.LIMITED_10
        case "prod" | "production":
            return Division.PRODUCTION
        case "ss" | "single stack":
            return Division.SINGLE_STACK
        case "rev" | "revo" | "revolver":
            return Division.REVOLVER
        case _ as unrecognized_value:
            _logger.debug("unknown division: %s", unrecognized_value)
            return Division.UNKNOWN


def parse_classification(s: str) -> Classification:
    match _sanitize_string(s).lower():
        case "gm" | "grandmaster" | "g":
            return Classification.GM
        case "m" | "master":
            return Classification.M
        case "a":
            return Classification.A
        case "b":
            return Classification.B
        case "c":
            return Classification.C
        case "d":
            return Classification.D
        case "u" | "x":
            return Classification.U
        case _ as unrecognized_value:
            _logger.debug("unknown classification: %s", unrecognized_value)
            return Classification.UNKNOWN


def parse_power_factor(s: str) -> PowerFactor:
    match _sanitize_string(s).lower():
        case "major":
            return PowerFactor.MAJOR
        case "minor":
            return PowerFactor.MINOR
        case _ as unrecognized_value:
            _logger.debug("unknown power factor: %s", unrecognized_value)
            return PowerFactor.UNKNOWN


def parse_member_number(s: str):
    return re.sub(r"[^0-9A-Z]", "", _sanitize_string(s).upper())


def parse_scoring(s: str):
    match _sanitize_string(s).lower():
        case "comstock":
            return Scoring.COMSTOCK
        case "virginia":
            return Scoring.VIRGINIA
        case "fixed":
            return Scoring.FIXED_TIME
        case "chrono":
            return Scoring.CHRONO
        case _ as unrecognized_value:
            _logger.debug("unknown scoring: %s", unrecognized_value)
            return Scoring.UNKNOWN


def parse_boolean(s: str):
    return _sanitize_string(s).lower() == "yes"



import logging
from enum import Enum, unique
from typing import List, Optional

from ...csv_utils import parse_csv_row, parse_float_value, parse_int_value
from ..fields import parse_boolean
from ..models import ParsedStageScore

_logger = logging.getLogger(__name__)


@unique
class StageScoreColumn(int, Enum):
    """Expected column indices"""

    STAGE_ID = 1
    SHOOTER_ID = 2
    DQ = 3
    DNF = 4
    A = 5
    B = 6
    C = 7
    D = 8
    M = 9
    NS = 10
    PROC = 11
    LATE_SHOT = 14
    EXTRA_SHOT = 15
    EXTRA_HIT = 16
    NPM = 17
    OTHER_PENALTY = 18
    PENALTY_POINTS = 19
    T1 = 20
    T2 = 21
    T3 = 22
    T4 = 23
    T5 = 24
    TIME = 25
    RAW_POINTS = 26
    TOTAL_POINTS = 27


def check_stage_score_columns(parsed_columns: List[str], fail_on_mismatch=False):
    check_failed = False
    expected_stage_id_col = parsed_columns[StageScoreColumn.STAGE_ID]
    if "stage" not in expected_stage_id_col.lower():
        _logger.error("CSV column order mismatch(?). Expected a stage id, but found: %s", expected_stage_id_col)
        check_failed = True
    # TODO: other checks, or just dynamically pick the values from the column name order?

    if fail_on_mismatch and check_failed:
        raise ValueError("expected parsed stage score column order mismatch detected")


def parse_match_report_stage_score_lines(
    stage_lines: List[str], stage_columns: Optional[List[str]] = None
) -> List[ParsedStageScore]:
    if stage_columns:
        check_stage_score_columns(stage_columns)
    stages: List[ParsedStageScore] = []
    for line in stage_lines:
        row = parse_csv_row(line)
        stage = ParsedStageScore(
            # Integer attributes
            stage_id=parse_int_value(row[StageScoreColumn.STAGE_ID].strip()),
            competitor_id=parse_int_value(row[StageScoreColumn.SHOOTER_ID].strip()),
            a=parse_int_value(row[StageScoreColumn.A].strip()) or 0,
            b=parse_int_value(row[StageScoreColumn.B].strip()) or 0,
            c=parse_int_value(row[StageScoreColumn.C].strip()) or 0,
            d=parse_int_value(row[StageScoreColumn.D].strip()) or 0,
            m=parse_int_value(row[StageScoreColumn.M].strip()) or 0,
            npm=parse_int_value(row[StageScoreColumn.NPM].strip()) or 0,
            ns=parse_int_value(row[StageScoreColumn.NS].strip()) or 0,
            procedural=parse_int_value(row[StageScoreColumn.PROC].strip()) or 0,
            late_shot=parse_int_value(row[StageScoreColumn.LATE_SHOT].strip()) or 0,
            extra_shot=parse_int_value(row[StageScoreColumn.EXTRA_SHOT].strip()) or 0,
            extra_hit=parse_int_value(row[StageScoreColumn.EXTRA_HIT].strip()) or 0,
            other_penalty=parse_int_value(row[StageScoreColumn.OTHER_PENALTY].strip()) or 0,
            # Float attributes
            t1=parse_float_value(row[StageScoreColumn.T1].strip()) or 0.0,
            t2=parse_float_value(row[StageScoreColumn.T2].strip()) or 0.0,
            t3=parse_float_value(row[StageScoreColumn.T3].strip()) or 0.0,
            t4=parse_float_value(row[StageScoreColumn.T4].strip()) or 0.0,
            t5=parse_float_value(row[StageScoreColumn.T5].strip()) or 0.0,
            time=parse_float_value(row[StageScoreColumn.TIME].strip()) or 0.0,
            # Boolean attributes
            dq=parse_boolean(row[StageScoreColumn.DQ].strip()),
            dnf=parse_boolean(row[StageScoreColumn.DNF].strip()),
        )
        stages.append(stage)
    return stages



import logging
import re
from enum import Enum, unique
from typing import List, Optional

from ....enums import Scoring
from ...csv_utils import parse_csv_row, parse_int_value
from ..fields import parse_scoring
from ..models import ParsedStage

_logger = logging.getLogger(__name__)


@unique
class StageColumn(int, Enum):
    ID = 0
    MIN_ROUNDS = 2
    MAX_POINTS = 3
    CLASSIFIER = 4
    CLASSIFIER_NUM = 5
    STAGE_NAME = 6
    SCORING = 7


def check_stage_columns(parsed_columns: List[str], fail_on_mismatch=False):
    check_failed = False
    expected_stage_name_col = parsed_columns[StageColumn.STAGE_NAME]
    if "name" not in expected_stage_name_col.lower():
        _logger.error("CSV column order mismatch(?). Expected a stage name, but found: %s", expected_stage_name_col)
        check_failed = True
    # TODO: other checks, or just dynamically pick the values from the column name order?

    if fail_on_mismatch and check_failed:
        raise ValueError("expected parsed stage column order mismatch detected")


def parse_match_report_stage_column_lines(column_lines: List[str]) -> Optional[List[str]]:
    n_header_cols = len(column_lines)
    if n_header_cols == 0:
        _logger.debug("no lines found for stage column names")
        return None
    if n_header_cols > 1:
        _logger.warning("expected only one header line, but found %d", n_header_cols)
    columns: Optional[List[str]] = parse_csv_row(column_lines[0])
    return columns


def parse_match_report_stage_lines(
    stage_lines: List[str], stage_columns: Optional[List[str]] = None
) -> List[ParsedStage]:
    if stage_columns:
        check_stage_columns(stage_columns)
    stages: List[ParsedStage] = []
    for line in stage_lines:
        row = parse_csv_row(line)
        stage = ParsedStage(
            internal_id=int(re.sub(r"[^0-9]", "", row[StageColumn.ID].strip())),
            name=row[StageColumn.STAGE_NAME].strip(),
            min_rounds=parse_int_value(row[StageColumn.MIN_ROUNDS]),
            max_points=parse_int_value(row[StageColumn.MAX_POINTS]),
            classifier=row[StageColumn.CLASSIFIER].strip().lower() == "yes",
            classifier_number=row[StageColumn.CLASSIFIER_NUM].strip(),
            scoring_type=parse_scoring(row[StageColumn.SCORING])
            if len(row) > StageColumn.SCORING
            else Scoring.COMSTOCK,
        )
        stages.append(stage)
    return stages



import logging
from dataclasses import dataclass, field
from typing import List

from ..constants import MatchReportSectionPrefix

_logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ParsedMatchLines:
    info: List[str] = field(default_factory=list)
    competitor_columns: List[str] = field(default_factory=list)
    competitor: List[str] = field(default_factory=list)
    stage_columns: List[str] = field(default_factory=list)
    stage: List[str] = field(default_factory=list)
    stage_score_columns: List[str] = field(default_factory=list)
    stage_score: List[str] = field(default_factory=list)


REPORT_LINE_PREFIX_INFO = f"{MatchReportSectionPrefix.INFO.value} "
REPORT_LINE_PREFIX_COLUMNS_COMPETITORS = f"{MatchReportSectionPrefix.COMPETITOR_HEADER.value} "
REPORT_LINE_PREFIX_COMPETITOR = f"{MatchReportSectionPrefix.COMPETITOR.value} "
REPORT_LINE_PREFIX_COLUMNS_STAGE = f"{MatchReportSectionPrefix.STAGE_HEADER.value} "
REPORT_LINE_PREFIX_STAGE = f"{MatchReportSectionPrefix.STAGE.value} "
REPORT_LINE_PREFIX_COLUMNS_STAGE_SCORE = f"{MatchReportSectionPrefix.STAGE_SCORE_HEADER.value} "
REPORT_LINE_PREFIX_STAGE_SCORE = f"{MatchReportSectionPrefix.STAGE_SCORE.value} "


def parse_match_report_response_lines(response: str) -> ParsedMatchLines:
    """
    First step in parsing involves categorizing sections of the response text based on each line's prefix
    """
    info_lines: List[str] = []
    competitor_column_lines: List[str] = []
    competitor_lines: List[str] = []
    stage_lines: List[str] = []
    stage_score_column_lines: List[str] = []
    stage_score_lines: List[str] = []
    stage_column_lines: List[str] = []
    response_lines = response.splitlines()

    for line in response_lines:
        sanitized_line = line.strip()
        if sanitized_line.startswith(REPORT_LINE_PREFIX_INFO):
            info_lines.append(sanitized_line.partition(REPORT_LINE_PREFIX_INFO)[-1])
        elif sanitized_line.startswith(REPORT_LINE_PREFIX_COMPETITOR):
            competitor_lines.append(sanitized_line.partition(REPORT_LINE_PREFIX_COMPETITOR)[-1])
        elif sanitized_line.startswith(REPORT_LINE_PREFIX_STAGE):
            stage_lines.append(sanitized_line.partition(REPORT_LINE_PREFIX_STAGE)[-1])
        elif sanitized_line.startswith(REPORT_LINE_PREFIX_STAGE_SCORE):
            stage_score_lines.append(sanitized_line.partition(REPORT_LINE_PREFIX_STAGE_SCORE)[-1])
        elif sanitized_line.startswith(REPORT_LINE_PREFIX_COLUMNS_STAGE):
            stage_column_lines.append(sanitized_line.partition(REPORT_LINE_PREFIX_COLUMNS_STAGE)[-1])
        elif sanitized_line.startswith(REPORT_LINE_PREFIX_COLUMNS_COMPETITORS):
            competitor_column_lines.append(sanitized_line.partition(REPORT_LINE_PREFIX_COLUMNS_COMPETITORS)[-1])
        elif sanitized_line.startswith(REPORT_LINE_PREFIX_COLUMNS_STAGE_SCORE):
            stage_score_column_lines.append(sanitized_line.partition(REPORT_LINE_PREFIX_COLUMNS_STAGE_SCORE)[-1])
        else:
            _logger.warning("unrecognized line: %s", sanitized_line)

    return ParsedMatchLines(
        info=info_lines,
        stage_columns=stage_column_lines,
        stage=stage_lines,
        competitor_columns=competitor_column_lines,
        competitor=competitor_lines,
        stage_score_columns=stage_score_column_lines,
        stage_score=stage_score_lines,
    )



import logging
from enum import Enum, unique
from io import StringIO
from typing import Any, Callable, List, Mapping

import pandas as pd
from pandas._typing import FilePath, ReadCsvBuffer
from pandas.errors import EmptyDataError

from ..fields import parse_boolean, parse_power_factor
from ..models import ParsedStageScore

_logger = logging.getLogger(__name__)


@unique
class StageScoreColumnName(str, Enum):
    """Known competitor column names. May also includes dataframe references and custom columns added to the dataframe after parsing."""

    STAGE_ID = "Stage"
    COMPETITOR_ID = "Comp"
    DQ = "DQ"
    DNF = "DNF"
    A = "A"
    B = "B"
    C = "C"
    D = "D"
    M = "Miss"
    NS = "No Shoot"
    PROC = "Procedural"
    LATE_SHOT = "Late Shot"
    EXTRA_SHOT = "Extra Shot"
    EXTRA_HIT = "Extra Hit"
    NPM = "No Penalty Miss"
    OTHER_PENALTY = "Additional Penalty"
    PENALTY_POINTS = "Total Penalty"
    T1 = "T1"
    T2 = "T2"
    T3 = "T3"
    T4 = "T4"
    T5 = "T5"
    TIME = "Time"
    RAW_POINTS = "Raw Points"
    TOTAL_POINTS = "Total Points"
    STAGE_POWER_FACTOR = "Stage Power Factor"


CSV_CONVERTERS: Mapping[str, Callable[[str], Any]] = {
    StageScoreColumnName.DQ: parse_boolean,
    StageScoreColumnName.DNF: parse_boolean,
    StageScoreColumnName.STAGE_POWER_FACTOR: parse_power_factor,
}


def read_stage_scores_csv(filepath_or_buffer: FilePath | ReadCsvBuffer[bytes] | ReadCsvBuffer[str]):
    """Load a dataframe with stage scores as rows parsed from csv"""
    df = pd.read_csv(
        filepath_or_buffer,
        index_col=None,
        converters=CSV_CONVERTERS,
    )
    # df['hit_factor'] = (df['A'] * 5) # TODO
    return df


def parse_stage_scores(stage_scores_csv: str) -> List[ParsedStageScore]:
    """Parse CSV text into ParsedStageScore objects. Uses pandas for parsing."""
    try:
        df = read_stage_scores_csv(StringIO(stage_scores_csv))
    except EmptyDataError:
        _logger.error("failed to parse stage_scores csv into dataframe")
        return []

    stage_scores = [
        ParsedStageScore(
            stage_id=stage_id,
            competitor_id=competitor_id,
            a=a,
            b=b,
            c=c,
            d=d,
            m=m,
            npm=npm,
            ns=ns,
            procedural=procedural,
            late_shot=late_shot,
            extra_shot=extra_shot,
            extra_hit=extra_hit,
            other_penalty=other_penalty,
            t1=t1,
            t2=t2,
            t3=t3,
            t4=t4,
            t5=t5,
            time=time,
            dq=dq,
            dnf=dnf,
        )
        for (
            stage_id,
            competitor_id,
            a,
            b,
            c,
            d,
            m,
            npm,
            ns,
            procedural,
            late_shot,
            extra_shot,
            extra_hit,
            other_penalty,
            t1,
            t2,
            t3,
            t4,
            t5,
            time,
            dq,
            dnf,
        ) in zip(
            df[StageScoreColumnName.STAGE_ID],
            df[StageScoreColumnName.COMPETITOR_ID],
            df[StageScoreColumnName.A],
            df[StageScoreColumnName.B],
            df[StageScoreColumnName.C],
            df[StageScoreColumnName.D],
            df[StageScoreColumnName.M],
            df[StageScoreColumnName.NPM],
            df[StageScoreColumnName.NS],
            df[StageScoreColumnName.PROC],
            df[StageScoreColumnName.LATE_SHOT],
            df[StageScoreColumnName.EXTRA_SHOT],
            df[StageScoreColumnName.EXTRA_HIT],
            df[StageScoreColumnName.OTHER_PENALTY],
            df[StageScoreColumnName.T1],
            df[StageScoreColumnName.T2],
            df[StageScoreColumnName.T3],
            df[StageScoreColumnName.T4],
            df[StageScoreColumnName.T5],
            df[StageScoreColumnName.TIME],
            df[StageScoreColumnName.DQ],
            df[StageScoreColumnName.DNF],
        )
    ]
    return stage_scores



import csv
from typing import Any, Callable, Optional, TypeVar, Union

TCastReturn = TypeVar("TCastReturn")


def _safe_cast(
    value: Any, caster: Callable[[Any], TCastReturn], default: Optional[TCastReturn] = None
) -> Optional[TCastReturn]:
    try:
        return caster(value)
    except (ValueError, TypeError):
        return default


def parse_csv_row(row_string: str):
    """
    Parse a string containing one row of CSV data
    """
    values = list(csv.reader([row_string]))
    return values[0] if values else None


def parse_int_value(v: Optional[Union[str, int]]) -> Optional[int]:
    if isinstance(v, str) and v.isdigit():
        return _safe_cast(v, int)
    if isinstance(v, (int, float, complex)) and not isinstance(v, bool):
        return _safe_cast(v, int)
    return None


def parse_float_value(v: str | float | None) -> Optional[float]:
    if isinstance(v, str):
        return _safe_cast(v, float)
    if isinstance(v, (int, float, complex)) and not isinstance(v, bool):
        return _safe_cast(v, float)
    return None



import httpx

_hitfactorpy_CACHE_DIR = "../../.hitfactorpy_cache"
_test_match_id = "01e65294-8692-4adf-9897-29dc01a68360"


def get_cached_report(cache_dir: str = _hitfactorpy_CACHE_DIR, match_id: str = _test_match_id):
    with open(f"{cache_dir}/reports/match__{match_id}.txt", "r") as f:
        return f.read()


def get_match_report(match_id: str = _test_match_id):
    """WIP"""
    if match_id and match_id is not _test_match_id:
        match_url = f"https://practiscore.com/reports/web/{match_id}"
        response = httpx.get(match_url)
        return response.text

    # for development
    get_cached_report(match_id)



import logging
from enum import Enum, unique
from io import StringIO
from typing import Any, Callable, List, Mapping

import pandas as pd
from pandas._typing import FilePath, ReadCsvBuffer
from pandas.errors import EmptyDataError

from ..fields import parse_boolean, parse_scoring
from ..models import ParsedStage

_logger = logging.getLogger(__name__)


@unique
class StageColumnName(str, Enum):
    """Known competitor column names. May also include dataframe references and custom columns added to the dataframe after parsing."""

    ID = "index"  # ex: used to reference `df.index`
    MIN_ROUNDS = "Minimum Rounds"
    MAX_POINTS = "Maximum Points"
    CLASSIFIER = "Classifier"
    CLASSIFIER_NUM = "Classifier_No"
    STAGE_NAME = "Stage_name"
    SCORING = "ScoringType"


CSV_CONVERTERS: Mapping[str, Callable[[str], Any]] = {
    StageColumnName.SCORING: parse_scoring,
    StageColumnName.CLASSIFIER: parse_boolean,
}


def read_stages_csv(filepath_or_buffer: FilePath | ReadCsvBuffer[bytes] | ReadCsvBuffer[str]):
    """Load a dataframe with stages parsed from csv"""
    df = pd.read_csv(
        filepath_or_buffer,
        index_col="Number",
        converters=CSV_CONVERTERS,
    )
    return df


def parse_stages(stage_csv_text: str) -> List[ParsedStage]:
    """Parse CSV text into ParsedStage objects. Uses pandas for parsing."""
    try:
        df = read_stages_csv(StringIO(stage_csv_text))
    except EmptyDataError:
        _logger.error("failed to parse stages csv into dataframe")
        return []

    stages = [
        ParsedStage(
            internal_id=internal_id,
            name=stage_name,
            scoring_type=scoring,
            min_rounds=min_rounds,
            max_points=max_points,
            classifier=classifier,
            classifier_number=classifier_number,
        )
        for internal_id, stage_name, scoring, min_rounds, max_points, classifier, classifier_number in zip(
            df.index,
            df[StageColumnName.STAGE_NAME],
            df[StageColumnName.SCORING],
            df[StageColumnName.MIN_ROUNDS],
            df[StageColumnName.MAX_POINTS],
            df[StageColumnName.CLASSIFIER],
            df[StageColumnName.CLASSIFIER_NUM],
        )
    ]
    # TODO: how to handle this edge case?
    #             scoring_type=parse_scoring(row[StageColumn.SCORING])
    #             if len(row) > StageColumn.SCORING
    #             else Scoring.COMSTOCK,
    return stages



from enum import Enum, unique


@unique
class Scoring(str, Enum):
    COMSTOCK = "comstock"
    VIRGINIA = "virginia"
    FIXED_TIME = "fixedTime"
    CHRONO = "chrono"
    UNKNOWN = "unknown"


@unique
class Division(str, Enum):
    PCC = "pcc"
    OPEN = "open"
    LIMITED = "limited"
    CARRY_OPTICS = "carryOptics"
    LIMITED_10 = "limited10"
    PRODUCTION = "production"
    SINGLE_STACK = "singleStack"
    REVOLVER = "revolver"

    """
    Unknown (missing data or parsing failed)
    """
    UNKNOWN = "unknown"


@unique
class PowerFactor(str, Enum):
    MAJOR = "major"
    MINOR = "minor"

    """
    Unknown (missing data or parsing failed)
    """
    UNKNOWN = "unknown"


@unique
class Classification(str, Enum):
    GM = "GM"
    M = "M"
    A = "A"
    B = "B"
    C = "C"
    D = "D"

    """
    Unclassified
    """
    U = "U"

    """
    Unknown (missing data or parsing failed)
    """
    UNKNOWN = "unknown"


@unique
class MatchLevel(int, Enum):
    I = 1  # noqa: E741
    II = 2
    III = 3
    IV = 4



from datetime import datetime
from typing import List, Optional

from pydantic.dataclasses import dataclass

from ...enums import Classification, Division, MatchLevel, PowerFactor, Scoring


@dataclass(frozen=True)
class ParsedCompetitor:
    """Competitor info parsed from match report"""

    internal_id: int
    member_number: Optional[str] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    division: Optional[Division] = None
    classification: Optional[Classification] = None
    power_factor: Optional[PowerFactor] = None
    dq: bool = False
    reentry: bool = False


@dataclass(frozen=True)
class ParsedStage:
    """Stage info parsed from match report"""

    internal_id: int
    name: Optional[str] = None
    min_rounds: Optional[int] = 0
    max_points: Optional[int] = 0
    classifier: bool = False
    classifier_number: Optional[str] = None
    scoring_type: Scoring = Scoring.COMSTOCK


@dataclass(frozen=True)
class ParsedStageScore:
    """Stage score parsed from match report"""

    competitor_id: Optional[int] = None
    stage_id: Optional[int] = None
    a: int = 0
    b: int = 0
    c: int = 0
    d: int = 0
    m: int = 0
    npm: int = 0
    ns: int = 0
    procedural: int = 0
    late_shot: int = 0
    extra_shot: int = 0
    extra_hit: int = 0
    other_penalty: int = 0
    t1: float = 0.0
    t2: float = 0.0
    t3: float = 0.0
    t4: float = 0.0
    t5: float = 0.0
    time: float = 0.0
    dq: bool = False
    dnf: bool = False


@dataclass(frozen=True)
class ParsedMatchInfo:
    """Match info parsed from header of match report"""

    name: Optional[str] = None
    raw_date: Optional[str] = None
    date: Optional[datetime] = None
    match_level: Optional[MatchLevel] = None


@dataclass(frozen=True)
class ParsedMatchReport(ParsedMatchInfo):
    """All data that could be parsed from the match report"""

    competitors: Optional[List[ParsedCompetitor]] = None
    stages: Optional[List[ParsedStage]] = None
    stage_scores: Optional[List[ParsedStageScore]] = None



import logging
from collections import defaultdict
from dataclasses import dataclass

from ..constants import MatchReportSectionPrefix

_logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ParsedMatchReportSections:
    """Report split into sections by prefix (prefix is stripped)"""

    info: str
    competitors: str
    stages: str
    stage_scores: str


def parse_match_report_sections(response: str) -> ParsedMatchReportSections:
    """
    First step in parsing involves categorizing sections of the response text based on each line's prefix
    """
    response_lines = response.splitlines()
    categorized_lines = defaultdict(list)

    for line in response_lines:
        if not line.startswith("$END"):
            try:
                prefix, line_content = line.split(None, 1)
                categorized_lines[prefix].append(line_content)
            except ValueError:
                pass

    if not (competitor_columns := categorized_lines.get(MatchReportSectionPrefix.COMPETITOR_HEADER)):
        _logger.warning("failed to parse competitor columns section")
        competitor_columns = [""]
    if not (stage_columns := categorized_lines.get(MatchReportSectionPrefix.STAGE_HEADER)):
        _logger.warning("failed to parse stage columns section")
        stage_columns = [""]
    if not (stage_score_columns := categorized_lines.get(MatchReportSectionPrefix.STAGE_SCORE_HEADER)):
        _logger.warning("failed to parse stage_score columns section")
        stage_score_columns = [""]

    return ParsedMatchReportSections(
        info="\n".join(categorized_lines.get(MatchReportSectionPrefix.INFO, [])),
        competitors=f"{competitor_columns[0]}\n"
        + "\n".join(categorized_lines.get(MatchReportSectionPrefix.COMPETITOR, [])),
        stages=f"{stage_columns[0]}\n" + "\n".join(categorized_lines.get(MatchReportSectionPrefix.STAGE, [])),
        stage_scores=f"{stage_score_columns[0]}\n"
        + "\n".join(categorized_lines.get(MatchReportSectionPrefix.STAGE_SCORE, [])),
    )



import json
from enum import Enum, unique
from pathlib import Path
from typing import Any, NamedTuple, Optional

import typer
from pydantic import schema_json_of
from pydantic.json import pydantic_encoder

CLI_NAME = "hitfactorpy"

cli = typer.Typer()


@unique
class MatchParserOption(str, Enum):
    PANDAS = "pandas"
    STRICT = "strict"


@cli.command()
def parse_match(
    match_report_file: Path, json_indent: Optional[int] = None, parser: MatchParserOption = MatchParserOption.PANDAS
):
    """
    Parse a match report file into JSON
    """
    if parser == MatchParserOption.PANDAS:
        from hitfactorpy.parsers.match_report.pandas import parse_match_report
    else:
        from hitfactorpy.parsers.match_report.strict import parse_match_report  # type:ignore

    with open(match_report_file, "r") as f:
        report_text = f.read()

    report = parse_match_report(report_text)
    typer.echo(json.dumps(report, indent=json_indent, default=pydantic_encoder))


@unique
class JsonSchemaModelArgument(str, Enum):
    MATCH_REPORT = "match"
    COMPETITOR = "competitor"
    STAGE = "stage"
    STAGE_SCORE = "stage-score"


@cli.command()
def json_schema(model: JsonSchemaModelArgument = JsonSchemaModelArgument.MATCH_REPORT, title: Optional[str] = None):
    """
    Dump the JSON Schema for a model
    """
    from hitfactorpy.parsers.match_report import models

    class ModelConfig(NamedTuple):
        """A model and its default title"""

        model: Any
        title: str

    models_config = {
        JsonSchemaModelArgument.MATCH_REPORT: ModelConfig(models.ParsedMatchReport, "MatchReport Schema"),
        JsonSchemaModelArgument.COMPETITOR: ModelConfig(models.ParsedCompetitor, "Competitor Schema"),
        JsonSchemaModelArgument.STAGE: ModelConfig(models.ParsedStage, "Stage schema"),
        JsonSchemaModelArgument.STAGE_SCORE: ModelConfig(models.ParsedStageScore, "StageScore Schema"),
    }
    config = models_config[model]

    typer.echo(schema_json_of(config.model, title=title or config.title))


if __name__ == "__main__":
    cli()



from enum import Enum, unique


@unique
class MatchReportSectionPrefix(str, Enum):
    PS = "$PRACTISCORE"
    INFO = "$INFO"
    COMPETITOR_HEADER = "D"
    COMPETITOR = "E"
    STAGE_HEADER = "F"
    STAGE = "G"
    STAGE_SCORE_HEADER = "H"
    STAGE_SCORE = "I"


@unique
class MatchReportInfoSectionFields(str, Enum):
    """Fields in the $INFO section that can be parsed via a single split on ':'"""

    MATCH_NAME = "Match name"
    MATCH_DATE = "Match date"
    MATCH_LEVEL = "Match Level"
    REGION = "Region"
    CLUB_NAME = "Club Name"
    CLUB_CODE = "Club Code"
    PS_PRODUCT = "PractiScore_Product"
    PS_VERSION = "PractiScore_Version"



import logging
import re
from enum import Enum, unique
from typing import List, Optional

from ...csv_utils import parse_csv_row
from ..fields import parse_classification, parse_division, parse_member_number, parse_power_factor
from ..models import ParsedCompetitor

_logger = logging.getLogger(__name__)


@unique
class CompetitorColumn(int, Enum):
    """Expected column indices"""

    ID = 0
    MEMBER_NUM = 1
    FIRST_NAME = 2
    LAST_NAME = 3
    DQ_PISTOL = 4
    DQ_RIFLE = 5
    DQ_SHOTGUN = 6
    REENTRY = 7
    CLASS = 8
    DIVISION = 9
    POWER_FACTOR = 12


def check_competitor_columns(parsed_columns: List[str], fail_on_mismatch=False):
    check_failed = False
    expected_first_name_col = parsed_columns[CompetitorColumn.FIRST_NAME]
    if "first" not in expected_first_name_col.lower():
        _logger.error("CSV column order mismatch(?). Expected 'FirstName', but found: %s", expected_first_name_col)
        check_failed = True
    # TODO: other checks, or just dynamically pick the values from the column name order?

    if fail_on_mismatch and check_failed:
        raise ValueError("expected parsed column order mismatch detected")


def _is_dq(competitor_row: List[str]) -> bool:
    return "yes" in [
        competitor_row[CompetitorColumn.DQ_PISTOL].lower(),
        competitor_row[CompetitorColumn.DQ_RIFLE].lower(),
        competitor_row[CompetitorColumn.DQ_SHOTGUN].lower(),
    ]


def parse_match_report_competitor_column_lines(column_lines: List[str]) -> Optional[List[str]]:
    n_header_cols = len(column_lines)
    if n_header_cols == 0:
        _logger.debug("no lines found for column names")
        return None
    if n_header_cols > 1:
        _logger.error("expected only one header line, but found %d", n_header_cols)
    columns: Optional[List[str]] = parse_csv_row(column_lines[0])
    return columns


def parse_match_report_competitor_lines(
    competitor_lines: List[str], competitor_columns: Optional[List[str]] = None
) -> List[ParsedCompetitor]:
    if competitor_columns:
        check_competitor_columns(competitor_columns)
    competitors: List[ParsedCompetitor] = []
    for line in competitor_lines:
        row = parse_csv_row(line)
        competitor = ParsedCompetitor(
            internal_id=int(re.sub(r"[^0-9]", "", row[CompetitorColumn.ID].strip())),
            member_number=parse_member_number(row[CompetitorColumn.MEMBER_NUM]),
            first_name=row[CompetitorColumn.FIRST_NAME].strip(),
            last_name=row[CompetitorColumn.LAST_NAME].strip(),
            division=parse_division(row[CompetitorColumn.DIVISION]),
            classification=parse_classification(row[CompetitorColumn.CLASS]),
            power_factor=parse_power_factor(row[CompetitorColumn.POWER_FACTOR]),
            reentry=row[CompetitorColumn.REENTRY].lower() == "yes",
            dq=_is_dq(row),
        )
        competitors.append(competitor)
    return competitors



from ..match_info import parse_match_info
from ..models import ParsedMatchReport
from .competitor import parse_match_report_competitor_column_lines, parse_match_report_competitor_lines
from .report_lines import parse_match_report_response_lines
from .stage import parse_match_report_stage_column_lines, parse_match_report_stage_lines
from .stage_score import parse_match_report_stage_score_lines


def parse_match_report(report_text: str) -> ParsedMatchReport:
    report_lines = parse_match_report_response_lines(report_text)
    match_info = parse_match_info(report_lines.info)
    competitor_columns = parse_match_report_competitor_column_lines(report_lines.competitor_columns)
    competitors = parse_match_report_competitor_lines(report_lines.competitor, competitor_columns)
    stage_columns = parse_match_report_stage_column_lines(report_lines.stage_columns)
    stages = parse_match_report_stage_lines(report_lines.stage, stage_columns)
    stage_scores = parse_match_report_stage_score_lines(report_lines.stage_score)
    return ParsedMatchReport(
        name=match_info.name,
        raw_date=match_info.raw_date,
        date=match_info.date,
        match_level=match_info.match_level,
        competitors=competitors,
        stages=stages,
        stage_scores=stage_scores,
    )



from io import StringIO
from typing import NamedTuple

from pandas import DataFrame

from .competitor import read_competitors_csv
from .report_sections import parse_match_report_sections
from .stage import read_stages_csv
from .stage_score import read_stage_scores_csv


class MatchReportDataframes(NamedTuple):
    competitors: DataFrame
    stages: DataFrame
    stage_scores: DataFrame


def load_match_report_dataframes(report_text: str) -> MatchReportDataframes:
    """Parse each CSV section of the match report into dataframes"""
    sections = parse_match_report_sections(report_text)
    competitors = read_competitors_csv(StringIO(sections.competitors))
    stages = read_stages_csv(StringIO(sections.stages))
    stage_scores = read_stage_scores_csv(StringIO(sections.stage_scores))

    return MatchReportDataframes(competitors, stages, stage_scores)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hitfactorpy',
 'hitfactorpy.parsers',
 'hitfactorpy.parsers.match_report',
 'hitfactorpy.parsers.match_report.pandas',
 'hitfactorpy.parsers.match_report.strict']

package_data = \
{'': ['*']}

install_requires = \
['httpx[http2]>=0.23.3,<0.24.0',
 'pandas>=1.5.2,<2.0.0',
 'pydantic[email]>=1.10.4,<2.0.0',
 'typer[all]>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['hitfactorpy = hitfactorpy.cli:cli']}

setup_kwargs = {
    'name': 'hitfactorpy',
    'version': '0.0.4',
    'description': 'Python tools for parsing and analyzing practical match reports',
    'long_description': "# hitfactorpy\n\n[![Main](https://github.com/cahna/hitfactorpy/actions/workflows/main.yaml/badge.svg)](https://github.com/cahna/hitfactorpy/actions/workflows/main.yaml)\n\nPython tools for parsing and analyzing practical match reports.\n\n## Status\n\n**Work in progress...**\n\nDocumentation website: [https:/cahna.github.io/hitfactorpy](https:/cahna.github.io/hitfactorpy)\n\nCurrently only supports USPSA match reports.\n\n## Library usage\n\nSee parsers available in [`hitfactorpy.parsers.match_report`](https://github.com/cahna/hitfactorpy/tree/main/hitfactorpy/parsers/match_report) and examples in [repository's `tests` directory](https://github.com/cahna/hitfactorpy/tree/main/tests).\n\n## CLI Usage\n\nAfter installing, run the `hitfactorpy` command should be added to `$PATH`.\n\n```console\n$ hitfactorpy --help\n```\n\n[Complete CLI documentation](https://cahna.github.io/hitfactorpy/)\n",
    'author': 'Conor Heine',
    'author_email': 'conor.heine@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<3.12',
}


setup(**setup_kwargs)



from ..match_info import parse_match_info
from ..models import ParsedMatchReport
from .competitor import parse_competitors
from .report_sections import parse_match_report_sections
from .stage import parse_stages
from .stage_score import parse_stage_scores


def parse_match_report(report_text: str) -> ParsedMatchReport:
    sections = parse_match_report_sections(report_text)
    match_info = parse_match_info(sections.info.split("\n"))
    competitors = parse_competitors(sections.competitors)
    stages = parse_stages(sections.stages)
    stage_scores = parse_stage_scores(sections.stage_scores)
    return ParsedMatchReport(
        name=match_info.name,
        raw_date=match_info.raw_date,
        date=match_info.date,
        match_level=match_info.match_level,
        competitors=competitors,
        stages=stages,
        stage_scores=stage_scores,
    )



import logging
from enum import Enum, unique
from io import StringIO
from typing import Any, Callable, List, Mapping

import pandas as pd
from pandas._typing import FilePath, ReadCsvBuffer
from pandas.errors import EmptyDataError

from ..fields import parse_boolean, parse_classification, parse_division, parse_member_number, parse_power_factor
from ..models import ParsedCompetitor

_logger = logging.getLogger(__name__)


@unique
class CompetitorColumnName(str, Enum):
    """Known competitor column names. Also includes dataframe references and custom columns added to the dataframe."""

    ID = "index"
    MEMBER_NUM = "USPSA"
    FIRST_NAME = "FirstName"
    LAST_NAME = "LastName"
    DQ_PISTOL = "DQPistol"
    DQ_RIFLE = "DQRifle"
    DQ_SHOTGUN = "DQShotgun"
    REENTRY = "Reentry"
    CLASS = "Class"
    DIVISION = "Division"
    POWER_FACTOR = "Power Factor"

    # Below this line are custom columns added to the pandas dataframe
    IS_DQ = "is_dq"


CONVERTERS: Mapping[str, Callable[[str], Any]] = {
    CompetitorColumnName.MEMBER_NUM: parse_member_number,
    CompetitorColumnName.DIVISION: parse_division,
    CompetitorColumnName.CLASS: parse_classification,
    CompetitorColumnName.POWER_FACTOR: parse_power_factor,
    CompetitorColumnName.REENTRY: parse_boolean,
    CompetitorColumnName.DQ_PISTOL: parse_boolean,
    CompetitorColumnName.DQ_RIFLE: parse_boolean,
    CompetitorColumnName.DQ_SHOTGUN: parse_boolean,
}


def read_competitors_csv(filepath_or_buffer: FilePath | ReadCsvBuffer[bytes] | ReadCsvBuffer[str]):
    """Load a dataframe with competitor rows parsed from csv"""
    df = pd.read_csv(
        filepath_or_buffer,
        index_col="Comp",
        converters=CONVERTERS,
    )
    df[CompetitorColumnName.IS_DQ] = (
        df[CompetitorColumnName.DQ_PISTOL] | df[CompetitorColumnName.DQ_RIFLE] | df[CompetitorColumnName.DQ_SHOTGUN]
    )
    return df


def parse_competitors(competitor_csv_text: str) -> List[ParsedCompetitor]:
    """Parse CSV text into ParsedCompetitor objects. Uses pandas for parsing."""
    try:
        df = read_competitors_csv(StringIO(competitor_csv_text))
    except EmptyDataError:
        _logger.error("failed to parse competitors csv into dataframe")
        return []

    competitors = [
        ParsedCompetitor(
            internal_id=internal_id,
            member_number=member_number,
            first_name=first_name,
            last_name=last_name,
            division=division,
            classification=classification,
            power_factor=power_factor,
            reentry=reentry,
            dq=dq,
        )
        for internal_id, member_number, first_name, last_name, classification, division, power_factor, reentry, dq in zip(
            df.index,
            df[CompetitorColumnName.MEMBER_NUM],
            df[CompetitorColumnName.FIRST_NAME],
            df[CompetitorColumnName.LAST_NAME],
            df[CompetitorColumnName.CLASS],
            df[CompetitorColumnName.DIVISION],
            df[CompetitorColumnName.POWER_FACTOR],
            df[CompetitorColumnName.REENTRY],
            df[CompetitorColumnName.IS_DQ],
        )
    ]

    return competitors
